Require Import Coq.Logic.FunctionalExtensionality.
Require Import Coq.Program.Equality.

Require Export Metalib.Metatheory.
Require Export Metalib.LibLNgen.

Require Export Definitions.

(** NOTE: Auxiliary theorems are hidden in generated documentation.
    In general, there is a [_rec] version of every lemma involving
    [open] and [close]. *)


(* *********************************************************************** *)
(** * Induction principles for nonterminals *)

Scheme term_ind' := Induction for term Sort Prop.

Definition term_mutind :=
  fun H1 H2 H3 H4 H5 =>
  term_ind' H1 H2 H3 H4 H5.

Scheme term_rec' := Induction for term Sort Type.

Definition term_mutrec :=
  fun H1 H2 H3 H4 H5 =>
  term_rec' H1 H2 H3 H4 H5.


(* *********************************************************************** *)
(** * Close *)

Fixpoint close_rec (n : nat) (x : atom) (t : term) : term :=
  match t with
  | BVar m => if lt_ge_dec m n then BVar m else BVar (S m)
  | Var y => if (x == y) then BVar n else Var y
  | Let t1 t2 => Let (close_rec n x t1) (close_rec (S n) x t2)
  | Lam t => Lam (close_rec (S n) x t)
  | App t1 t2 => App (close_rec n x t1) (close_rec n x t2)
  | Unit => Unit
  | Pair t1 t2 => Pair (close_rec n x t1) (close_rec n x t2)
  | Fst t => Fst (close_rec n x t)
  | Snd t => Snd (close_rec n x t)
  | InjL t => InjL (close_rec n x t)
  | InjR t => InjR (close_rec n x t)
  | Match t t1 t2 => Match (close_rec n x t) (close_rec (S n) x t1) (close_rec (S n) x t2)
  end.

Definition close x t := close_rec 0 x t.


(* *********************************************************************** *)
(** * Size *)

Fixpoint size (t : term) : nat :=
  match t with
  | BVar _ => 1
  | Var _ => 1
  | Let t1 t2 => 1 + size t1 + size t2
  | Lam t => 1 + size t
  | Unit => 1
  | App t1 t2 => 1 + size t1 + size t2
  | Pair t1 t2 => 1 + size t1 + size t2
  | Fst t => 1 + size t
  | Snd t => 1 + size t
  | InjL t => 1 + size t
  | InjR t => 1 + size t
  | Match t t1 t2 => 1 + size t + size t1 + size t2
  end.


(* *********************************************************************** *)
(** * Degree *)

(** These define only an upper bound, not a strict upper bound. *)

Inductive degree : nat -> term -> Prop :=
| degree_Var : forall n x,
    degree n (Var x)
| degree_BVar : forall n1 n2,
    lt n2 n1 ->
    degree n1 (BVar n2)
| degree_Let : forall n t1 t2,
    degree n t1 ->
    degree (S n) t2 ->
    degree n (Let t1 t2)
| degree_Lam : forall n t,
    degree (S n) t ->
    degree n (Lam t)
| degree_App : forall n t1 t2,
    degree n t1 ->
    degree n t2 ->
    degree n (App t1 t2)
| degree_Unit : forall n,
    degree n Unit
| degree_Pair : forall n t1 t2,
    degree n t1 ->
    degree n t2 ->
    degree n (Pair t1 t2)
| degree_Fst : forall n t,
    degree n t ->
    degree n (Fst t)
| degree_Snd : forall n t,
    degree n t ->
    degree n (Snd t)
| degree_InjL : forall n t,
    degree n t ->
    degree n (InjL t)
| degree_InjR : forall n t,
    degree n t ->
    degree n (InjR t)
| degree_Match : forall n t t1 t2,
    degree n t ->
    degree (S n) t1 ->
    degree (S n) t2 ->
    degree n (Match t t1 t2)
.

Scheme degree_ind' := Induction for degree Sort Prop.

Definition degree_mutind :=
  fun H1 H2 H3 H4 H5 =>
  degree_ind' H1 H2 H3 H4 H5.

Hint Constructors degree : core lngen.


(* *********************************************************************** *)
(** * Local closure (version in [Type], induction principles) *)

Inductive lc_set : term -> Type :=
| lc_set_Var : forall x,
    lc_set (Var x)
| lc_set_Let : forall t1 t2,
    lc_set t1 ->
    (forall x : atom, lc_set (open t2 (Var x))) ->
    lc_set (Let t1 t2)
| lc_set_Lam : forall t,
    (forall x : atom, lc_set (open t (Var x))) ->
    lc_set (Lam t)
| lc_set_App : forall t1 t2,
    lc_set t1 ->
    lc_set t2 ->
    lc_set (App t1 t2)
| lc_set_Unit :
    lc_set Unit
| lc_set_Pair : forall t1 t2,
    lc_set t1 ->
    lc_set t2 ->
    lc_set (Pair t1 t2)
| lc_set_Fst : forall t,
    lc_set t ->
    lc_set (Fst t)
| lc_set_Snd : forall t,
    lc_set t ->
    lc_set (Snd t)
| lc_set_InjL : forall t,
    lc_set t ->
    lc_set (InjL t)
| lc_set_InjR : forall t,
    lc_set t ->
    lc_set (InjR t)
| lc_set_Match : forall t t1 t2,
    lc_set t ->
    (forall x : atom, lc_set (open t1 (Var x))) ->
    (forall x : atom, lc_set (open t2 (Var x))) ->
    lc_set (Match t t1 t2)
.

Scheme lc_ind' := Induction for lc Sort Prop.

Definition lc_mutind :=
  fun H1 H2 H3 H4 =>
  lc_ind' H1 H2 H3 H4.

Scheme lc_set_ind' := Induction for lc_set Sort Prop.

Definition lc_set_mutind :=
  fun H1 H2 H3 H4 =>
  lc_set_ind' H1 H2 H3 H4.

Scheme lc_set_rec' := Induction for lc_set Sort Type.

Definition lc_set_mutrec :=
  fun H1 H2 H3 H4 =>
  lc_set_rec' H1 H2 H3 H4.

Hint Constructors lc : core lngen.

Hint Constructors lc_set : core lngen.


(* *********************************************************************** *)
(** * Body *)

Definition body t := forall x, lc (open t (Var x)).

Hint Unfold body : lngen.


(* *********************************************************************** *)
(** * Tactic support *)

(** Additional hint declarations. *)

Hint Resolve @plus_le_compat : lngen.

(** Redefine some tactics. *)

Ltac default_case_split ::=
  first
    [ progress destruct_notin
    | progress destruct_sum
    | progress safe_f_equal
    ].


(* *********************************************************************** *)
(** * Theorems about [size] *)

Ltac default_auto ::= auto with arith lngen; tauto.
Ltac default_autorewrite ::= fail.

(* begin hide *)

Lemma size_min_mutual :
(forall t, 1 <= size t).
Proof.
apply_mutual_ind term_mutind;
default_simp.
Qed.

(* end hide *)

Lemma size_min :
forall t, 1 <= size t.
Proof.
pose proof size_min_mutual as H; intuition eauto.
Qed.

Hint Resolve size_min : lngen.

(* begin hide *)

Lemma size_close_rec_mutual :
(forall t x n,
  size (close_rec n x t) = size t).
Proof.
apply_mutual_ind term_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_close_rec :
forall t x n,
  size (close_rec n x t) = size t.
Proof.
pose proof size_close_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_close_rec : lngen.
Hint Rewrite size_close_rec using solve [auto] : lngen.

(* end hide *)

Lemma size_close :
forall t x,
  size (close x t) = size t.
Proof.
unfold close; default_simp.
Qed.

Hint Resolve size_close : lngen.
Hint Rewrite size_close using solve [auto] : lngen.

(* begin hide *)

Lemma size_open_rec_mutual :
(forall t1 t2 n,
  size t1 <= size (open_rec n t2 t1)).
Proof.
apply_mutual_ind term_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_open_rec :
forall t1 t2 n,
  size t1 <= size (open_rec n t2 t1).
Proof.
pose proof size_open_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_open_rec : lngen.

(* end hide *)

Lemma size_open :
forall t1 t2,
  size t1 <= size (open t1 t2).
Proof.
unfold open; default_simp.
Qed.

Hint Resolve size_open : lngen.

(* begin hide *)

Lemma size_open_rec_var_mutual :
(forall t x n,
  size (open_rec n (Var x) t) = size t).
Proof.
apply_mutual_ind term_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_open_rec_var :
forall t x n,
  size (open_rec n (Var x) t) = size t.
Proof.
pose proof size_open_rec_var_mutual as H; intuition eauto.
Qed.

Hint Resolve size_open_rec_var : lngen.
Hint Rewrite size_open_rec_var using solve [auto] : lngen.

(* end hide *)

Lemma size_open_var :
forall t x,
  size (open t (Var x)) = size t.
Proof.
unfold open; default_simp.
Qed.

Hint Resolve size_open_var : lngen.
Hint Rewrite size_open_var using solve [auto] : lngen.


(* *********************************************************************** *)
(** * Theorems about [degree] *)

Ltac default_auto ::= auto with lngen; tauto.
Ltac default_autorewrite ::= fail.

(* begin hide *)

Lemma degree_S_mutual :
(forall n e,
  degree n e ->
  degree (S n) e).
Proof.
apply_mutual_ind degree_mutind;
default_simp.
Qed.

(* end hide *)

Lemma degree_S :
forall n e,
  degree n e ->
  degree (S n) e.
Proof.
pose proof degree_S_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_S : lngen.

Lemma degree_O :
forall n e,
  degree O e ->
  degree n e.
Proof.
induction n; default_simp.
Qed.

Hint Resolve degree_O : lngen.

(* begin hide *)

Lemma degree_close_rec_mutual :
(forall e x n,
  degree n e ->
  degree (S n) (close_rec n x e)).
Proof.
apply_mutual_ind term_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_close_rec :
forall e x n,
  degree n e ->
  degree (S n) (close_rec n x e).
Proof.
pose proof degree_close_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_close_rec : lngen.

(* end hide *)

Lemma degree_close :
forall e x,
  degree 0 e ->
  degree 1 (close x e).
Proof.
unfold close; default_simp.
Qed.

Hint Resolve degree_close : lngen.

(* begin hide *)

Lemma degree_close_rec_inv_mutual :
(forall e x n,
  degree (S n) (close_rec n x e) ->
  degree n e).
Proof.
apply_mutual_ind term_mutind;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_close_rec_inv :
forall e x n,
  degree (S n) (close_rec n x e) ->
  degree n e.
Proof.
pose proof degree_close_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_close_rec_inv : lngen.

(* end hide *)

Lemma degree_close_inv :
forall t1 x1,
  degree 1 (close x1 t1) ->
  degree 0 t1.
Proof.
unfold close; eauto with lngen.
Qed.

Hint Immediate degree_close_inv : lngen.

(* begin hide *)

Lemma degree_open_rec_mutual :
(forall t1 t2 n1,
  degree (S n1) t1 ->
  degree n1 t2 ->
  degree n1 (open_rec n1 t2 t1)).
Proof.
apply_mutual_ind term_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_open_rec :
forall t1 t2 n1,
  degree (S n1) t1 ->
  degree n1 t2 ->
  degree n1 (open_rec n1 t2 t1).
Proof.
pose proof degree_open_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_open_rec : lngen.

(* end hide *)

Lemma degree_open :
forall t1 t2,
  degree 1 t1 ->
  degree 0 t2 ->
  degree 0 (open t1 t2).
Proof.
unfold open; default_simp.
Qed.

Hint Resolve degree_open : lngen.

(* begin hide *)

Lemma degree_open_rec_inv_mutual :
(forall t1 t2 n1,
  degree n1 (open_rec n1 t2 t1) ->
  degree (S n1) t1).
Proof.
apply_mutual_ind term_mutind;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_open_rec_inv :
forall t1 t2 n1,
  degree n1 (open_rec n1 t2 t1) ->
  degree (S n1) t1.
Proof.
pose proof degree_open_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_open_rec_inv : lngen.

(* end hide *)

Lemma degree_open_inv :
forall t1 t2,
  degree 0 (open t1 t2) ->
  degree 1 t1.
Proof.
unfold open; eauto with lngen.
Qed.

Hint Immediate degree_open_inv : lngen.


(* *********************************************************************** *)
(** * Theorems about [open] and [close] *)

Ltac default_auto ::= auto with lngen brute_force; tauto.
Ltac default_autorewrite ::= fail.

(* begin hide *)

Lemma close_rec_inj_mutual :
(forall t1 t2 x1 n1,
  close_rec n1 x1 t1 = close_rec n1 x1 t2 ->
  t1 = t2).
Proof.
apply_mutual_ind term_mutind;
intros; match goal with
          | |- _ = ?term => destruct term
        end;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_rec_inj :
forall t1 t2 x1 n1,
  close_rec n1 x1 t1 = close_rec n1 x1 t2 ->
  t1 = t2.
Proof.
pose proof close_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate close_rec_inj : lngen.

(* end hide *)

Lemma close_inj :
forall t1 t2 x1,
  close x1 t1 = close x1 t2 ->
  t1 = t2.
Proof.
unfold close; eauto with lngen.
Qed.

Hint Immediate close_inj : lngen.

(* begin hide *)

Lemma close_rec_open_rec_mutual :
(forall t1 x1 n1,
  x1 `notin` fv t1 ->
  close_rec n1 x1 (open_rec n1 (Var x1) t1) = t1).
Proof.
apply_mutual_ind term_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_rec_open_rec :
forall t1 x1 n1,
  x1 `notin` fv t1 ->
  close_rec n1 x1 (open_rec n1 (Var x1) t1) = t1.
Proof.
pose proof close_rec_open_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve close_rec_open_rec : lngen.
Hint Rewrite close_rec_open_rec using solve [auto] : lngen.

(* end hide *)

Lemma close_open :
forall t1 x1,
  x1 `notin` fv t1 ->
  close x1 (open t1 (Var x1)) = t1.
Proof.
unfold close; unfold open; default_simp.
Qed.

Hint Resolve close_open : lngen.
Hint Rewrite close_open using solve [auto] : lngen.

(* begin hide *)

Lemma open_rec_close_rec_mutual :
(forall t1 x1 n1,
  open_rec n1 (Var x1) (close_rec n1 x1 t1) = t1).
Proof.
apply_mutual_ind term_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_rec_close_rec :
forall t1 x1 n1,
  open_rec n1 (Var x1) (close_rec n1 x1 t1) = t1.
Proof.
pose proof open_rec_close_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve open_rec_close_rec : lngen.
Hint Rewrite open_rec_close_rec using solve [auto] : lngen.

(* end hide *)

Lemma open_close :
forall t1 x1,
  open (close x1 t1) (Var x1) = t1.
Proof.
unfold close; unfold open; default_simp.
Qed.

Hint Resolve open_close : lngen.
Hint Rewrite open_close using solve [auto] : lngen.

(* begin hide *)

Lemma open_rec_inj_mutual :
(forall t2 t1 x1 n1,
  x1 `notin` fv t2 ->
  x1 `notin` fv t1 ->
  open_rec n1 (Var x1) t2 = open_rec n1 (Var x1) t1 ->
  t2 = t1).
Proof.
apply_mutual_ind term_mutind;
intros; match goal with
          | |- _ = ?term => destruct term
        end;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_rec_inj :
forall t2 t1 x1 n1,
  x1 `notin` fv t2 ->
  x1 `notin` fv t1 ->
  open_rec n1 (Var x1) t2 = open_rec n1 (Var x1) t1 ->
  t2 = t1.
Proof.
pose proof open_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate open_rec_inj : lngen.

(* end hide *)

Lemma open_inj :
forall t2 t1 x1,
  x1 `notin` fv t2 ->
  x1 `notin` fv t1 ->
  open t2 (Var x1) = open t1 (Var x1) ->
  t2 = t1.
Proof.
unfold open; eauto with lngen.
Qed.

Hint Immediate open_inj : lngen.


(* *********************************************************************** *)
(** * Theorems about [lc] *)

Ltac default_auto ::= auto with lngen brute_force; tauto.
Ltac default_autorewrite ::= autorewrite with lngen.

(* begin hide *)

Lemma degree_of_lc_mutual :
(forall t1,
  lc t1 ->
  degree 0 t1).
Proof.
apply_mutual_ind lc_mutind;
intros;
let x1 := fresh "x1" in pick_fresh x1;
repeat (match goal with
          | H1 : _, H2 : _ |- _ => specialize H1 with H2
        end);
default_simp; eauto with lngen.
Qed.

(* end hide *)

Lemma degree_of_lc :
forall t1,
  lc t1 ->
  degree 0 t1.
Proof.
pose proof degree_of_lc_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_of_lc : lngen.

(* begin hide *)

Lemma lc_of_degree_size_mutual :
forall i1,
(forall t1,
  size t1 = i1 ->
  degree 0 t1 ->
  lc t1).
Proof.
intros i1; pattern i1; apply lt_wf_rec;
clear i1; intros i1 H1;
apply_mutual_ind term_mutind;
default_simp;
(* non-trivial cases *)
constructor; default_simp; eapply_first_lt_hyp;
(* instantiate the size *)
match goal with
  | |- _ = _ => reflexivity
  | _ => idtac
end;
instantiate;
(* everything should be easy now *)
default_simp.
Qed.

(* end hide *)

Lemma lc_of_degree :
forall t1,
  degree 0 t1 ->
  lc t1.
Proof.
intros t1; intros;
pose proof (lc_of_degree_size_mutual (size t1));
intuition eauto.
Qed.

Hint Resolve lc_of_degree : lngen.

Ltac typ_lc_exists_tac :=
  repeat (match goal with
            | H : _ |- _ =>
              fail 1
          end).

Ltac term_lc_exists_tac :=
  repeat (match goal with
            | H : _ |- _ =>
              let J1 := fresh in pose proof H as J1; apply degree_of_lc in J1; clear H
          end).

Lemma lc_Lam_exists :
forall x1 t1,
  lc (open t1 (Var x1)) ->
  lc (Lam t1).
Proof.
intros; term_lc_exists_tac; eauto with lngen.
Qed.

Hint Extern 1 (lc (Lam _)) =>
  let x1 := fresh in
  pick_fresh x1;
  apply (lc_Lam_exists x1) : lngen.

Lemma lc_Let_exists :
forall x t1 t2,
  lc t1 ->
  lc (open t2 (Var x)) ->
  lc (Let t1 t2).
Proof.
intros; term_lc_exists_tac; eauto with lngen.
Qed.

Hint Extern 1 (lc (Let _ _)) =>
  let x1 := fresh in
  pick_fresh x1;
  apply (lc_Let_exists x1) : lngen.

Lemma lc_Match_exists :
forall x1 x2 t t1 t2,
  lc t ->
  lc (open t1 (Var x1)) ->
  lc (open t2 (Var x2)) ->
  lc (Match t t1 t2).
Proof.
intros; term_lc_exists_tac; eauto with lngen.
Qed.

Hint Extern 1 (lc (Match _ _ _)) =>
  let x1 := fresh in
  let x2 := fresh in
  pick_fresh x1;
  pick_fresh x2;
  apply (lc_Match_exists x1 x2) : lngen.

Lemma lc_body :
forall t1 t2,
  body t1 ->
  lc t2 ->
  lc (open t1 t2).
Proof.
unfold body;
default_simp;
let x1 := fresh "x" in
pick_fresh x1;
specialize_all x1;
term_lc_exists_tac;
eauto with lngen.
Qed.

Hint Resolve lc_body : lngen.

Lemma lc_body_Lam_1 :
forall t1,
  lc (Lam t1) ->
  body t1.
Proof.
default_simp.
Qed.

Hint Resolve lc_body_Lam_1 : lngen.

Lemma lc_body_Let_2 :
forall t1 t2,
  lc (Let t1 t2) ->
  body t2.
Proof.
default_simp.
Qed.

Hint Resolve lc_body_Let_2 : lngen.

Lemma lc_body_Match_2 :
forall t t1 t2,
  lc (Match t t1 t2) ->
  body t1.
Proof.
default_simp.
Qed.

Hint Resolve lc_body_Match_2 : lngen.

Lemma lc_body_Match_3 :
forall t1 t2 t3,
  lc (Match t1 t2 t3) ->
  body t3.
Proof.
default_simp.
Qed.

Hint Resolve lc_body_Match_3 : lngen.


(* begin hide *)

Lemma lc_unique_mutual :
(forall t1 (proof2 proof3 : lc t1), proof2 = proof3).
Proof.
apply_mutual_ind lc_mutind;
intros;
let proof1 := fresh "proof1" in
rename_last_into proof1; dependent destruction proof1;
f_equal; default_simp; auto using @functional_extensionality_dep with lngen.
Qed.

(* end hide *)

Lemma lc_unique :
forall t1 (proof2 proof3 : lc t1), proof2 = proof3.
Proof.
pose proof lc_unique_mutual as H; intuition eauto.
Qed.

Hint Resolve lc_unique : lngen.

(* begin hide *)

Lemma lc_of_lc_set_mutual :
(forall t1, lc_set t1 -> lc t1).
Proof.
apply_mutual_ind lc_set_mutind;
default_simp.
Qed.

(* end hide *)

Lemma lc_of_lc_set :
forall t1, lc_set t1 -> lc t1.
Proof.
pose proof lc_of_lc_set_mutual as H; intuition eauto.
Qed.

Hint Resolve lc_of_lc_set : lngen.

(* begin hide *)

Lemma lc_set_of_lc_size_mutual :
forall i1,
(forall t1,
  size t1 = i1 ->
  lc t1 ->
  lc_set t1).
Proof.
intros i1; pattern i1; apply lt_wf_rec;
clear i1; intros i1 H1;
apply_mutual_ind term_mutrec;
default_simp;
try solve [assert False by default_simp; tauto];
(* non-trivial cases *)
constructor; default_simp;
try first [apply lc_set_of_lc];
default_simp; eapply_first_lt_hyp;
(* instantiate the size *)
match goal with
  | |- _ = _ => reflexivity
  | _ => idtac
end;
instantiate;
(* everything should be easy now *)
default_simp.
Qed.

(* end hide *)

Lemma lc_set_of_lc :
forall t1,
  lc t1 ->
  lc_set t1.
Proof.
intros t1; intros;
pose proof (lc_set_of_lc_size_mutual (size t1));
intuition eauto.
Qed.

Hint Resolve lc_set_of_lc : lngen.


(* *********************************************************************** *)
(** * More theorems about [open] and [close] *)

Ltac default_auto ::= auto with lngen; tauto.
Ltac default_autorewrite ::= fail.

(* begin hide *)

Lemma close_rec_degree_mutual :
(forall t1 x1 n1,
  degree n1 t1 ->
  x1 `notin` fv t1 ->
  close_rec n1 x1 t1 = t1).
Proof.
apply_mutual_ind term_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_rec_degree :
forall t1 x1 n1,
  degree n1 t1 ->
  x1 `notin` fv t1 ->
  close_rec n1 x1 t1 = t1.
Proof.
pose proof close_rec_degree_mutual as H; intuition eauto.
Qed.

Hint Resolve close_rec_degree : lngen.
Hint Rewrite close_rec_degree using solve [auto] : lngen.

(* end hide *)

Lemma close_lc :
forall t1 x1,
  lc t1 ->
  x1 `notin` fv t1 ->
  close x1 t1 = t1.
Proof.
unfold close; default_simp.
Qed.

Hint Resolve close_lc : lngen.
Hint Rewrite close_lc using solve [auto] : lngen.

(* begin hide *)

Lemma open_rec_degree_mutual :
(forall t2 t1 n1,
  degree n1 t2 ->
  open_rec n1 t1 t2 = t2).
Proof.
apply_mutual_ind term_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_rec_degree :
forall t2 t1 n1,
  degree n1 t2 ->
  open_rec n1 t1 t2 = t2.
Proof.
pose proof open_rec_degree_mutual as H; intuition eauto.
Qed.

Hint Resolve open_rec_degree : lngen.
Hint Rewrite open_rec_degree using solve [auto] : lngen.

(* end hide *)

Lemma open_lc :
forall t2 t1,
  lc t2 ->
  open t2 t1 = t2.
Proof.
unfold open; default_simp.
Qed.

Hint Resolve open_lc : lngen.
Hint Rewrite open_lc using solve [auto] : lngen.


(* *********************************************************************** *)
(** * Theorems about [fv] *)

Ltac default_auto ::= auto with set lngen; tauto.
Ltac default_autorewrite ::= autorewrite with lngen.

(* begin hide *)

Lemma fv_close_rec_mutual :
(forall t1 x1 n1,
  fv (close_rec n1 x1 t1) [=] remove x1 (fv t1)).
Proof.
apply_mutual_ind term_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_close_rec :
forall t1 x1 n1,
  fv (close_rec n1 x1 t1) [=] remove x1 (fv t1).
Proof.
pose proof fv_close_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_close_rec : lngen.
Hint Rewrite fv_close_rec using solve [auto] : lngen.

(* end hide *)

Lemma fv_close :
forall t1 x1,
  fv (close x1 t1) [=] remove x1 (fv t1).
Proof.
unfold close; default_simp.
Qed.

Hint Resolve fv_close : lngen.
Hint Rewrite fv_close using solve [auto] : lngen.

(* begin hide *)

Lemma fv_open_rec_lower_mutual :
(forall t1 t2 n1,
  fv t1 [<=] fv (open_rec n1 t2 t1)).
Proof.
apply_mutual_ind term_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_open_rec_lower :
forall t1 t2 n1,
  fv t1 [<=] fv (open_rec n1 t2 t1).
Proof.
pose proof fv_open_rec_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_open_rec_lower : lngen.

(* end hide *)

Lemma fv_open_lower :
forall t1 t2,
  fv t1 [<=] fv (open t1 t2).
Proof.
unfold open; default_simp.
Qed.

Hint Resolve fv_open_lower : lngen.

(* begin hide *)

Lemma fv_open_rec_upper_mutual :
(forall t1 t2 n1,
  fv (open_rec n1 t2 t1) [<=] fv t2 `union` fv t1).
Proof.
apply_mutual_ind term_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_open_rec_upper :
forall t1 t2 n1,
  fv (open_rec n1 t2 t1) [<=] fv t2 `union` fv t1.
Proof.
pose proof fv_open_rec_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_open_rec_upper : lngen.

(* end hide *)

Lemma fv_open_upper :
forall t1 t2,
  fv (open t1 t2) [<=] fv t2 `union` fv t1.
Proof.
unfold open; default_simp.
Qed.

Hint Resolve fv_open_upper : lngen.

(* begin hide *)

Lemma fv_subst_fresh_mutual :
(forall t1 t2 x1,
  x1 `notin` fv t1 ->
  fv (subst t2 x1 t1) [=] fv t1).
Proof.
apply_mutual_ind term_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

Lemma fv_subst_fresh :
forall t1 t2 x1,
  x1 `notin` fv t1 ->
  fv (subst t2 x1 t1) [=] fv t1.
Proof.
pose proof fv_subst_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_subst_fresh : lngen.
Hint Rewrite fv_subst_fresh using solve [auto] : lngen.

(* begin hide *)

Lemma fv_subst_lower_mutual :
(forall t1 t2 x1,
  remove x1 (fv t1) [<=] fv (subst t2 x1 t1)).
Proof.
apply_mutual_ind term_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

Lemma fv_subst_lower :
forall t1 t2 x1,
  remove x1 (fv t1) [<=] fv (subst t2 x1 t1).
Proof.
pose proof fv_subst_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_subst_lower : lngen.

(* begin hide *)

Lemma fv_subst_notin_mutual :
(forall t1 t2 x1 x2,
  x2 `notin` fv t1 ->
  x2 `notin` fv t2 ->
  x2 `notin` fv (subst t2 x1 t1)).
Proof.
apply_mutual_ind term_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

Lemma fv_subst_notin :
forall t1 t2 x1 x2,
  x2 `notin` fv t1 ->
  x2 `notin` fv t2 ->
  x2 `notin` fv (subst t2 x1 t1).
Proof.
pose proof fv_subst_notin_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_subst_notin : lngen.

(* begin hide *)

Lemma fv_subst_upper_mutual :
(forall t1 t2 x1,
  fv (subst t2 x1 t1) [<=] fv t2 `union` remove x1 (fv t1)).
Proof.
apply_mutual_ind term_mutind;
default_simp; fsetdec.
Qed.

(* end hide *)

Lemma fv_subst_upper :
forall t1 t2 x1,
  fv (subst t2 x1 t1) [<=] fv t2 `union` remove x1 (fv t1).
Proof.
pose proof fv_subst_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_subst_upper : lngen.


(* *********************************************************************** *)
(** * Theorems about [subst] *)

Ltac default_auto ::= auto with lngen brute_force; tauto.
Ltac default_autorewrite ::= autorewrite with lngen.

(* begin hide *)

Lemma subst_close_rec_mutual :
(forall t2 t1 x1 x2 n1,
  degree n1 t1 ->
  x1 <> x2 ->
  x2 `notin` fv t1 ->
  subst t1 x1 (close_rec n1 x2 t2) = close_rec n1 x2 (subst t1 x1 t2)).
Proof.
apply_mutual_ind term_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_close_rec :
forall t2 t1 x1 x2 n1,
  degree n1 t1 ->
  x1 <> x2 ->
  x2 `notin` fv t1 ->
  subst t1 x1 (close_rec n1 x2 t2) = close_rec n1 x2 (subst t1 x1 t2).
Proof.
pose proof subst_close_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_close_rec : lngen.

Lemma subst_close :
forall t2 t1 x1 x2,
  lc t1 ->  x1 <> x2 ->
  x2 `notin` fv t1 ->
  subst t1 x1 (close x2 t2) = close x2 (subst t1 x1 t2).
Proof.
unfold close; default_simp.
Qed.

Hint Resolve subst_close : lngen.

(* begin hide *)

Lemma subst_degree_mutual :
(forall t1 t2 x1 n1,
  degree n1 t1 ->
  degree n1 t2 ->
  degree n1 (subst t2 x1 t1)).
Proof.
apply_mutual_ind term_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_degree :
forall t1 t2 x1 n1,
  degree n1 t1 ->
  degree n1 t2 ->
  degree n1 (subst t2 x1 t1).
Proof.
pose proof subst_degree_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_degree : lngen.

(* begin hide *)

Lemma subst_fresh_eq_mutual :
(forall t2 t1 x1,
  x1 `notin` fv t2 ->
  subst t1 x1 t2 = t2).
Proof.
apply_mutual_ind term_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_fresh_eq :
forall t2 t1 x1,
  x1 `notin` fv t2 ->
  subst t1 x1 t2 = t2.
Proof.
pose proof subst_fresh_eq_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_fresh_eq : lngen.
Hint Rewrite subst_fresh_eq using solve [auto] : lngen.

(* begin hide *)

Lemma subst_fresh_same_mutual :
(forall t2 t1 x1,
  x1 `notin` fv t1 ->
  x1 `notin` fv (subst t1 x1 t2)).
Proof.
apply_mutual_ind term_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_fresh_same :
forall t2 t1 x1,
  x1 `notin` fv t1 ->
  x1 `notin` fv (subst t1 x1 t2).
Proof.
pose proof subst_fresh_same_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_fresh_same : lngen.

(* begin hide *)

Lemma subst_fresh_mutual :
(forall t2 t1 x1 x2,
  x1 `notin` fv t2 ->
  x1 `notin` fv t1 ->
  x1 `notin` fv (subst t1 x2 t2)).
Proof.
apply_mutual_ind term_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_fresh :
forall t2 t1 x1 x2,
  x1 `notin` fv t2 ->
  x1 `notin` fv t1 ->
  x1 `notin` fv (subst t1 x2 t2).
Proof.
pose proof subst_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_fresh : lngen.

Lemma subst_lc :
forall t1 t2 x1,
  lc t1 ->
  lc t2 ->
  lc (subst t2 x1 t1).
Proof.
default_simp.
Qed.

Hint Resolve subst_lc : lngen.

(* begin hide *)

Lemma subst_open_rec_mutual :
(forall e3 t1 t2 x1 n1,
  lc t1 ->
  subst t1 x1 (open_rec n1 t2 e3) = open_rec n1 (subst t1 x1 t2) (subst t1 x1 e3)).
Proof.
apply_mutual_ind term_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_open_rec :
forall e3 t1 t2 x1 n1,
  lc t1 ->
  subst t1 x1 (open_rec n1 t2 e3) = open_rec n1 (subst t1 x1 t2) (subst t1 x1 e3).
Proof.
pose proof subst_open_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_open_rec : lngen.

(* end hide *)

Lemma subst_open :
forall e3 t1 t2 x1,
  lc t1 ->
  subst t1 x1 (open e3 t2) = open (subst t1 x1 e3) (subst t1 x1 t2).
Proof.
unfold open; default_simp.
Qed.

Hint Resolve subst_open : lngen.

Lemma subst_open_var :
forall t2 t1 x1 x2,
  x1 <> x2 ->
  lc t1 ->
  open (subst t1 x1 t2) (Var x2) = subst t1 x1 (open t2 (Var x2)).
Proof.
intros; rewrite subst_open; default_simp.
Qed.

Hint Resolve subst_open_var : lngen.

(* begin hide *)

Lemma subst_spec_rec_mutual :
(forall t1 t2 x1 n1,
  subst t2 x1 t1 = open_rec n1 t2 (close_rec n1 x1 t1)).
Proof.
apply_mutual_ind term_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_spec_rec :
forall t1 t2 x1 n1,
  subst t2 x1 t1 = open_rec n1 t2 (close_rec n1 x1 t1).
Proof.
pose proof subst_spec_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_spec_rec : lngen.

(* end hide *)

Lemma subst_spec :
forall t1 t2 x1,
  subst t2 x1 t1 = open (close x1 t1) t2.
Proof.
unfold close; unfold open; default_simp.
Qed.

Hint Resolve subst_spec : lngen.

(* begin hide *)

Lemma subst_subst_mutual :
(forall t1 t2 e3 x2 x1,
  x2 `notin` fv t2 ->
  x2 <> x1 ->
  subst t2 x1 (subst e3 x2 t1) = subst (subst t2 x1 e3) x2 (subst t2 x1 t1)).
Proof.
apply_mutual_ind term_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_subst :
forall t1 t2 e3 x2 x1,
  x2 `notin` fv t2 ->
  x2 <> x1 ->
  subst t2 x1 (subst e3 x2 t1) = subst (subst t2 x1 e3) x2 (subst t2 x1 t1).
Proof.
pose proof subst_subst_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_subst : lngen.

(* begin hide *)

Lemma subst_close_rec_open_rec_mutual :
(forall t2 t1 x1 x2 n1,
  x2 `notin` fv t2 ->
  x2 `notin` fv t1 ->
  x2 <> x1 ->
  degree n1 t1 ->
  subst t1 x1 t2 = close_rec n1 x2 (subst t1 x1 (open_rec n1 (Var x2) t2))).
Proof.
apply_mutual_ind term_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_close_rec_open_rec :
forall t2 t1 x1 x2 n1,
  x2 `notin` fv t2 ->
  x2 `notin` fv t1 ->
  x2 <> x1 ->
  degree n1 t1 ->
  subst t1 x1 t2 = close_rec n1 x2 (subst t1 x1 (open_rec n1 (Var x2) t2)).
Proof.
pose proof subst_close_rec_open_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_close_rec_open_rec : lngen.

(* end hide *)

Lemma subst_close_open :
forall t2 t1 x1 x2,
  x2 `notin` fv t2 ->
  x2 `notin` fv t1 ->
  x2 <> x1 ->
  lc t1 ->
  subst t1 x1 t2 = close x2 (subst t1 x1 (open t2 (Var x2))).
Proof.
unfold close; unfold open; default_simp.
Qed.

Hint Resolve subst_close_open : lngen.

Lemma subst_Lam :
forall x2 t2 t1 x1,
  lc t1 ->
  x2 `notin` fv t1 `union` fv t2 `union` singleton x1 ->
  subst t1 x1 (Lam t2) = Lam (close x2 (subst t1 x1 (open t2 (Var x2)))).
Proof.
default_simp.
Qed.

Hint Resolve subst_Lam : lngen.

Lemma subst_Let :
forall t0 x2 t2 t1 x1,
  lc t0 ->
  lc t1 ->
  x2 `notin` fv t1 `union` fv t2 `union` singleton x1 ->
  subst t1 x1 (Let t0 t2) = Let (subst t1 x1 t0) (close x2 (subst t1 x1 (open t2 (Var x2)))).
Proof.
default_simp.
Qed.

Hint Resolve subst_Let : lngen.

Lemma subst_Match :
forall t0 x2 t2 t3 t1 x1,
  lc t0 ->
  lc t1 ->
  x2 `notin` fv t1 `union` fv t2 `union` fv t3 `union` singleton x1 ->
  subst t1 x1 (Match t0 t2 t3) = Match (subst t1 x1 t0) (close x2 (subst t1 x1 (open t2 (Var x2)))) (close x2 (subst t1 x1 (open t3 (Var x2)))).
Proof.
default_simp.
Qed.

Hint Resolve subst_Match : lngen.

(* begin hide *)

Lemma subst_intro_rec_mutual :
(forall t1 x1 t2 n1,
  x1 `notin` fv t1 ->
  open_rec n1 t2 t1 = subst t2 x1 (open_rec n1 (Var x1) t1)).
Proof.
apply_mutual_ind term_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_intro_rec :
forall t1 x1 t2 n1,
  x1 `notin` fv t1 ->
  open_rec n1 t2 t1 = subst t2 x1 (open_rec n1 (Var x1) t1).
Proof.
pose proof subst_intro_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_intro_rec : lngen.
Hint Rewrite subst_intro_rec using solve [auto] : lngen.

Lemma subst_intro :
forall x1 t1 t2,
  x1 `notin` fv t1 ->
  open t1 t2 = subst t2 x1 (open t1 (Var x1)).
Proof.
unfold open; default_simp.
Qed.

Hint Resolve subst_intro : lngen.


(* *********************************************************************** *)
(** * "Restore" tactics *)

Ltac default_auto ::= auto; tauto.
Ltac default_autorewrite ::= fail.
